"""
This file defines the OpinionatedParamsInterface, which is an interface for defining RAG-specific parameters.

The LastMileTracer class in the lastmile_tracer.py file inherits from the OpinionatedParamsImpl class,
which is the implementation of the OpinionatedParamsInterface defined in the opinionated_params_impl.py file.

The LastMileTracer API in the tracing.py file inherits from this OpinionatedParamsInterface.

This structure allows for a clear separation of the interface and its implementation, while providing
a convenient way to access the RAG-specific parameters through the LastMileTracer API.
"""

from enum import Enum
import abc
from typing import Any, Optional

from opentelemetry.trace.span import Span


class ParamKey(Enum):
    """
    Opinionated parameter keys for the RAG-specific parameters.
    """

    # High Level
    MODEL = "model"
    TEMPERATURE = "temperature"
    TOP_P = "top_p"
    CHUNK_SIZE = "chunk_size"
    TOP_K = "top_k"

    # Ingestion-specific
    CHUNK_STRATEGY = "chunk_strategy"

    # Query-specific
    DECOMPOSITION_STRATEGY = "decomposition_strategy"
    RETRIEVAL_MODEL = "retrieval_model"
    RERANKING_MODEL = "reranking_model"

    # Both
    EMBEDDING_MODEL = "embedding_model"
    EMBEDDING_DIMENSIONS = "embedding_dimensions"


class ManageParamsInterface(abc.ABC):
    """
    Interface for defining the RAG-specific Params.
    """

    @abc.abstractmethod
    def register_document_preprocess_params(
        self,
        chunk_size: Optional[int] = None,
        chunk_strategy: Optional[str] = None,
        extras: Optional[dict[str | ParamKey, Any]] = None,
        span: Optional[Span] = None,
    ) -> None:
        """
        Register the configuration of the document preprocessing step, like chunking.

        @param chunk_size (int): Chunk size
        @param chunk_strategy (str): Chunking algorithm (e.g. "sliding window")
        @param extras (dict[str | ParamKey, Any]): Any other params to also register
        @param span Optional(Span): The span to save the K-V pair in
            addition to regular paramSet. Defaults to
            `trace_api.get_current_span()` which is the most recent span
            generated by calling tracer.start_as_current_span
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def register_embedding_params(
        self,
        embedding_model: Optional[str] = None,
        embedding_dimensions: Optional[int] = None,
        extras: Optional[dict[str | ParamKey, Any]] = None,
        span: Optional[Span] = None,
    ) -> None:
        """
        Register the configuration of the embedding step.

        @params embedding_model: Embedding model
        @params embedding_dimensions: Cardinality of the embedding
        @param extras (dict[str | ParamKey, Any]): Any other params to also register
        @param span Optional(Span): The span to save the K-V pair in
            addition to regular paramSet. Defaults to
            `trace_api.get_current_span()` which is the most recent span
            generated by calling tracer.start_as_current_span
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def register_query_processing_params(
        self,
        embedding_model: Optional[str] = None,
        embedding_dimensions: Optional[int] = None,
        decomposition_strategy: Optional[str] = None,
        extras: Optional[dict[str | ParamKey, Any]] = None,
        span: Optional[Span] = None,
    ) -> None:
        """
        Register the configuration of the query processing step,
        such as query decomposition, embedding the query, etc.

        @param embedding_model (str): Embedding model used for the query
        @param embedding_dimensions (int): Cardinality of the embedding used for the query
        @param decomposition_strategy: Algorithm for how a query is processed
            (e.g. "generate subquestions", "query expansion", etc.)
        @param extras (dict[str | ParamKey, Any]): Any other params to also register
        @param span Optional(Span): The span to save the K-V pair in
            addition to regular paramSet. Defaults to
            `trace_api.get_current_span()` which is the most recent span
            generated by calling tracer.start_as_current_span
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def register_retrieval_params(
        self,
        top_k: Optional[int] = None,
        reranking_model: Optional[str] = None,
        extras: Optional[dict[str | ParamKey, Any]] = None,
        span: Optional[Span] = None,
    ) -> None:
        """
        Register the configuration of the retrieval step, such as top_k, etc.

        @param top_k (int): Number of embeddings to retrieve
        @param reranking_model (str): The model used for reranking results
        @param extras (dict[str | ParamKey, Any]): Any other params to also register
        @param span Optional(Span): The span to save the K-V pair in
            addition to regular paramSet. Defaults to
            `trace_api.get_current_span()` which is the most recent span
            generated by calling tracer.start_as_current_span
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def register_generation_params(
        self,
        model_params: dict[str | ParamKey, Any],
        extras: Optional[dict[str | ParamKey, Any]] = None,
        span: Optional[Span] = None,
    ) -> None:
        """
        Register the LLM settings like model, temperature, context_length used for the
        LLM generation phase of the RAG flow.

        Args:
        @param model_params (dict[str, Any]): The dictionary of model settings. Values must
            be JSON-serializable
            NOTE: It is strongly recommended for this dictionary to comply with the OpenAI API spec. The params will automatically exclude messages from the ParamSet.
        @param extras (dict[str | ParamKey, Any]): Any other params to also register
        @param span Optional(Span): The span to save the K-V pair in
            addition to regular paramSet. Defaults to
            `trace_api.get_current_span()` which is the most recent span
            generated by calling tracer.start_as_current_span
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def get_params(self) -> dict[str, Any]:
        """
        Returns the params_dict that contains all the parameters that have been
        registered with a trace so far.

        If this is called outside of an active trace, it will return the
        global params dict that contains K-V pairs that will be common in all
        future traces (if they haven't been cleared via `clear_params()`)
        """
        raise NotImplementedError("Not implemented directly, this is an API")

    @abc.abstractmethod
    def register_param(
        self,
        key: str | ParamKey,
        value: Any,
        span: Optional[Span] = None,
    ) -> None:
        """
        Define the parameter K-V pair to save for the current trace instance.

        If this is called outside of an active trace, it will save this into
        a global params dict that contains K-V pairs that will be common in all
        future traces (if they haven't been cleared via `clear_params()`)

        @param key (str | ParamKey): The name of the parameter to be saved
        @param value (Any): The value of the parameter to be saved
        @param span Optional(Span): The span to save the K-V pair in
            addition to regular paramSet. This can be helpful for debugging
            when going through the trace. Defaults to
            `trace_api.get_current_span()` which is the most recent span
            generated by calling tracer.start_as_current_span
        """
        raise NotImplementedError("Not implemented directly, this is an API")

    @abc.abstractmethod
    def register_params(
        self,
        params: dict[str | ParamKey, Any],
        should_overwrite: bool = False,
        span: Optional[Span] = None,
    ) -> None:
        """
        Helper function for individually calling `register_param`, with the
        added capability of clearing existing parameters if they exist.

        If this is called outside of an active trace, it will save these into
        a global params dict that contains K-V pairs that will be common in all
        future traces (if they haven't been cleared via `clear_params()`)

        @param params dict[str | ParamKey, Any]: The parameter K-V pairs to save
        @param should_overwrite (bool): Whether to clear existing parameters
            if they already exist. Defaults to false.
        @param span Optional(Span): The span to save the K-V pair in
            addition to regular paramSet. This can be helpful for debugging
            when going through the trace. Defaults to
            `trace_api.get_current_span()` which is the most recent span
            generated by calling tracer.start_as_current_span

        Define the parameter K-V pair to save for the current trace instance
        """
        raise NotImplementedError("Not implemented directly, this is an API")

    @abc.abstractmethod
    def clear_params(
        self,
        should_clear_global_params: bool = False,
    ) -> None:
        """
        Clearing all existing parameters for the current trace instance.

        If this is called outside of an active trace, it will only clear the
        global params dict if `should_clear_global_params` is set to True.

        @param should_clear_global_params (bool): Whether to clear the global
        K-V pairs in addition to the current trace params. Defaults to false
        """
        raise NotImplementedError("Not implemented directly, this is an API")
